from sphinx.domains import Domain, ObjType
from sphinx.locale import l_
from sphinx.directives import ObjectDescription
from sphinx.util.docfields import GroupedField, Field, TypedField
from sphinx.roles import XRefRole
from sphinx.util.nodes import make_refnode
from sphinx import addnodes
from sphinx.domains.python import _pseudo_parse_arglist


class PHPObject(ObjectDescription):
    required_arguments = 1
    
    def get_signatures(self):
        tmp= ObjectDescription.get_signatures(self)
        return tmp
    
    def add_target_and_index(self, name, sig, signode):
        if name not in self.state.document.ids:
            signode['names'].append(name)
            signode['ids'].append(name)
            signode['first'] = (not self.names)
            self.state.document.note_explicit_target(signode)
            objects = self.env.domaindata['php']['objects']
            if name in objects:
                self.env.warn(
                    self.env.docname,
                    'duplicate object description of %s, ' % name +
                    'other instance in ' +
                    self.env.doc2path(objects[name][0]),
                    self.lineno)
            objects[name] = (self.env.docname, self.objtype)
            text = self.get_index_display_string(name)
            self.indexnode['entries'].append(('single', text,
                                              name, name))

class PHPNamespacelike(PHPObject):
    doc_field_types = []
    
    def handle_signature(self, sig, signode):
        kw, name = sig.split()
        if "namespace" not in kw:
            raise ValueError
        namespace = name.split('\\')
        namespace_name = namespace[-1] 
        annotation = "namespace " + "\\".join(namespace[:-1])
        signode['fullname'] = namespace_name
        signode += addnodes.desc_annotation(annotation, annotation)
        signode += addnodes.desc_name(namespace_name, namespace_name)
        return "\\".join(namespace)
    
    def get_index_display_string(self, name):
        nsname = name.split("\\")[-1] 
        complement = " (%s" % (self.objtype,)
        if self.env.temp_data['php:namespace']:
            parent_ns = self.env.temp_data['php:namespace'].split("\\")[:-1]
            parent_ns = "\\".join(parent_ns)
            if parent_ns:
                complement += ", in namespace %s" % (parent_ns,)
        complement += ")"
        return "%s%s" % (nsname, complement)
    
    def before_content(self):
        # needed for automatic qualification of members (reset in subclasses)
        if self.names:
            self.env.temp_data['php:namespace'] = self.names[0]
            self.nsname_set = True

    def after_content(self):
        if self.clsname_set:
            self.env.temp_data['php:namespace'] = None
        
    

class PHPClasslike(PHPObject):
    doc_field_types = [
            Field('implements', label=l_('Implements'),
                       names=('impl', 'implement', 'implements'), has_arg=False,
                       rolename='php:ifc'),
            Field('extends', label=l_('Extends'),
                       names=('extends', 'extend'), has_arg=False,
                       rolename='class'),
    ]
    
    def handle_signature(self, sig, signode):
        if 'interface' in sig:
            annotation = "interface "
            class_name = sig.replace('interface', '', 1).strip()
        else:
            tokens = [el.strip() for el in sig.split('class')]
            if 'abstract' in tokens:
                annotation = 'abstract class '
            else:
                annotation = "class "
            class_name = tokens[1]
        if "php:namespace" in self.env.temp_data \
            and self.env.temp_data["php:namespace"]:
            annotation += "%s\\" % (self.env.temp_data["php:namespace"])

        signode['fullname'] = class_name
        signode += addnodes.desc_annotation(annotation, annotation)
        signode += addnodes.desc_name(class_name, class_name)
        return class_name
    
    def get_index_display_string(self, name):
        complement = " (%s" % (self.objtype)
        if "php:namespace" in self.env.temp_data \
            and self.env.temp_data["php:namespace"]:
            complement += ", in namespace %s" % (self.env.temp_data["php:namespace"])
        complement +=")"
        return "%s%s" % (name, complement)
    
#    def add_target_and_index(self, name, sig, signode):
#        if name not in self.state.document.ids:
#            signode['names'].append(name)
#            signode['ids'].append(name)
#            signode['first'] = (not self.names)
#            self.state.document.note_explicit_target(signode)
#            objects = self.env.domaindata['php']['objects']
#            if name in objects:
#                self.env.warn(
#                    self.env.docname,
#                    'duplicate object description of %s, ' % name +
#                    'other instance in ' +
#                    self.env.doc2path(objects[name][0]),
#                    self.lineno)
#            objects[name] = (self.env.docname, self.objtype)
#            self.indexnode['entries'].append(('single', "%s (%s)" % (name, self.objtype),
#                                              name, name))
    
    def before_content(self):
        # needed for automatic qualification of members (reset in subclasses)
        if self.names:
            self.env.temp_data['php:class'] = self.names[0]
            self.clsname_set = True

    def after_content(self):
        if self.clsname_set:
            self.env.temp_data['php:class'] = None
           
 

class PHPFunctionlike(PHPObject):
    doc_field_types = [
            Field('deprecated', label=l_('Deprecated'), rolename='obj',
                  names=("deprecated",), has_arg=False),
            GroupedField('since', label=l_('Changes'), rolename='obj',
                         names=("since",), can_collapse=True),
            TypedField('parameter', label=l_('Parameters'),
                       names=('param', 'parameter', 'arg', 'argument',
                              'keyword', 'kwarg', 'kwparam'),
                       typerolename='obj', typenames=('paramtype', 'type')),
            TypedField('variable', label=l_('Variables'), rolename='obj',
                       names=('var', 'ivar', 'cvar'),
                       typerolename='obj', typenames=('vartype',)),
            GroupedField('exceptions', label=l_('Raises'), rolename='exc',
                     names=('raises', 'raise', 'exception', 'except', 'throw', 'throws'),
                     can_collapse=True),
            Field('returnvalue', label=l_('Returns'), has_arg=False,
                  names=('returns', 'return')),
            Field('returntype', label=l_('Return type'), has_arg=False,
                  names=('rtype',)),
    ]
    
    def handle_signature(self, sig, signode):
        qualifiers, signature = sig.split('function')
        
        qualifiers = qualifiers.split()
        annotation = ""
        if "final" in qualifiers:
            annotation += "final "
        if "public" in qualifiers:
            annotation += "public "
        if "protected" in qualifiers:
            annotation += "protected "
        if "private" in qualifiers:
            annotation += "private "
        if "static" in qualifiers:
            annotation += "static "
        
        annotation += self.objtype + " "
            
        # fix if not prefix
        func_name, args = signature.strip().split("(")
        if "::" in func_name:
            fullname = func_name.split('::')[1]
            signode['class'] = func_name.split('::')[0]
            annotation += func_name.split('::')[0] + "::"
        else:
            cls_name = self.env.temp_data['php:class']
            if cls_name:
                fullname = func_name
                signode['class'] = cls_name
                annotation += cls_name.split('::')[0] + "::"
            else:
                fullname = func_name

        
        signode['fullname'] = fullname
        signode += addnodes.desc_annotation(annotation, annotation)
        signode += addnodes.desc_name(fullname , fullname)
        
        arglist = args.strip(")").strip()
        if not arglist:
            signode += addnodes.desc_parameterlist()
            return fullname
        
        _pseudo_parse_arglist(signode, arglist.strip(")"))
        return fullname
    
    def get_index_display_string(self, name):
        complement = " (%s" % (self.objtype)
        if "php:class" in self.env.temp_data \
            and self.env.temp_data["php:class"]:
            complement += ", in class %s" % (self.env.temp_data["php:class"])
        complement +=")"
        return "%s%s" % (name, complement)
    
#    def add_target_and_index(self, name, sig, signode):
#        if name not in self.state.document.ids:
#            signode['names'].append(name)
#            signode['ids'].append(name)
#            signode['first'] = (not self.names)
#            self.state.document.note_explicit_target(signode)
#            objects = self.env.domaindata['php']['objects']
#            if name in objects:
#                self.env.warn(
#                    self.env.docname,
#                    'duplicate object description of %s, ' % name +
#                    'other instance in ' +
#                    self.env.doc2path(objects[name][0]),
#                    self.lineno)
#            objects[name] = (self.env.docname, self.objtype)
#            if self.objtype=='method':
#                complement = "%s (in class %s)" \
#                                    % (name.split('::')[1], 
#                                       name.split('::')[0])
#            else:
#                complement = "%s (%s)" % (name, "function")
#            self.indexnode['entries'].append(('single', complement,
#                                              name, name)) 

class PHPVarlike(PHPObject):
    doc_field_types = []
    
    def handle_signature(self, sig, signode):
        qualifiers = sig.split()
        name = qualifiers.pop(-1).strip()
        
        annotation = ""
        if "final" in qualifiers:
            annotation += "final "
        if "public" in qualifiers:
            annotation += "public "
        if "protected" in qualifiers:
            annotation += "protected "
        if "private" in qualifiers:
            annotation += "private "
        if "static" in qualifiers:
            annotation += "static "
        
        annotation += self.objtype + " "
            
        if name.startswith("$"):
            name = name[1:]
        if "::" in name:
            fullname = name
            signode['class'] = name.split('::')[0]
        else:
            cls_name = self.env.temp_data.get('php:class', None)
            if cls_name:
                fullname = cls_name + "::" + name
                signode['class'] = cls_name
            else:
                fullname = "$" + name if self.objtype == "var" else name
        
        signode['fullname'] = fullname
        signode += addnodes.desc_annotation(annotation, annotation)
        signode += addnodes.desc_name(fullname , fullname)
#        arglist = args.strip(")").strip()
#        if not arglist:
#            signode += addnodes.desc_parameterlist()
#            return fullname
        
#        _pseudo_parse_arglist(signode, arglist.strip(")"))
        return fullname
    
    def get_index_display_string(self, name):
        complement = " (%s" % (self.objtype)
        if "php:class" in self.env.temp_data \
            and self.env.temp_data["php:class"]:
            complement += ", in class %s" % (self.env.temp_data["php:class"])
        complement +=")"
        return "%s%s" % (name, complement)
    
#    def add_target_and_index(self, name, sig, signode):
#        if name not in self.state.document.ids:
#            signode['names'].append(name)
#            signode['ids'].append(name)
#            signode['first'] = (not self.names)
#            self.state.document.note_explicit_target(signode)
#            objects = self.env.domaindata['php']['objects']
#            if name in objects:
#                self.env.warn(
#                    self.env.docname,
#                    'duplicate object description of %s, ' % name +
#                    'other instance in ' +
#                    self.env.doc2path(objects[name][0]),
#                    self.lineno)
#            objects[name] = (self.env.docname, self.objtype)
#            if '::' in name:
#                complement = "%s (%s in class %s)" \
#                                    % (name.split('::')[1],
#                                       self.objtype, 
#                                       name.split('::')[0])
#            else:
#                complement = "%s (%s)" % (name, "global")
#            self.indexnode['entries'].append(('single', complement,
#                                              name, name)) 


class PHPXRefRole(XRefRole):
    pass


class PHPDomain(Domain):
    """PHP language domain."""
    name = 'php'
    label = 'PHP'
    object_types = {
        'function':        ObjType(l_('function'),         'func', 'obj'),
        'var':             ObjType(l_('global variable'),  'var', 'obj'),
        'method':          ObjType(l_('method'),           'meth', 'obj'),
        'class':           ObjType(l_('class'),            'class', 'obj'),
        'const':           ObjType(l_('const'),            'const', 'obj'),
        'interface':       ObjType(l_('interface'),        'ifc', 'obj'),
        'exception':       ObjType(l_('exception'),        'exc', 'obj'),
        'attribute':       ObjType(l_('attribute'),        'attr', 'obj'),
    }

    directives = {
        'function':         PHPFunctionlike,
        'var':              PHPVarlike,
        'method':           PHPFunctionlike,
        'class':            PHPClasslike,
        'const':            PHPVarlike,
        'interface':        PHPClasslike,
        'exception':        PHPClasslike,
        'attribute':        PHPVarlike,
    }

    roles = {
        'func':  PHPXRefRole(fix_parens=False),
        'var':   PHPXRefRole(),
        'meth':  PHPXRefRole(fix_parens=False),
        'class': PHPXRefRole(),
        'const': PHPXRefRole(),
        'ifc':   PHPXRefRole(),
        'exc':   PHPXRefRole(),
        'attr':  PHPXRefRole(),

#        'obj':   PHPXRefRole(),

    }
    initial_data = {
        'objects': {},  # fullname -> docname, objtype
#        'modules': {},  # modname -> docname, synopsis, platform, deprecated
    }
    indices = []

    def clear_doc(self, docname):
        for fullname, (fn, _) in self.data['objects'].items():
            if fn == docname:
                del self.data['objects'][fullname]
#        for modname, (fn, _, _, _) in self.data['modules'].items():
#            if fn == docname:
#                del self.data['modules'][modname]

    def resolve_xref(self, env, fromdocname, builder,
                     typ, target, node, contnode):
        if target in self.data["objects"]:
            return make_refnode(builder, fromdocname, self.data["objects"][target][0], target,
                                contnode, target)
        else:
            return None

    def get_objects(self):
#        for modname, info in self.data['modules'].iteritems():
#            yield (modname, modname, 'module', info[0], 'module-' + modname, 0)
        for refname, (docname, type) in self.data['objects'].iteritems():
            yield (refname, refname, type, docname, refname, 1)

def setup(app):
    app.add_domain(PHPDomain)